#!/usr/bin/env python
# Written by Alan Milligan and Danica Sutherland (2023-)
# Based on CPSC 540 Julia code by Mark Schmidt
# and some CPSC 340 Python code by Mike Gelbart and Nam Hee Kim, among others

import os
from pathlib import Path

import numpy as np
import matplotlib.pyplot as plt


# make sure we're working in the directory this file lives in,
# for simplicity with imports and relative paths
os.chdir(Path(__file__).parent.resolve())

# question code
from utils import (
    test_and_plot,
    load_dataset,
    load_mnist,
    main,
    handle,
    run,
)
from naive_bayes import NaiveBayes, VQNB
from neural_net import NeuralNetRegressor


def sample_cat(n_sample, pdf, rng=None):
    if rng is None:
        rng = np.random.default_rng()
    raise NotImplementedError()


@handle("expval-mc")
def expval_mc():
    max_n = 100_000
    pdf = (0.41, 0.17, 0.42)
    n_repeats = 3

    fig, ax = plt.subplots()
    ax.set_xscale("log")
    ax.set_xlabel("$t$")
    ax.set_ylabel("Monte Carlo approximation")
    ax.set_xlim(1, max_n)
    ax.set_ylim(-1.5, 1.5)

    raise NotImplementedError()

    fn = Path("..") / "figs" / "expval_game.pdf"
    fig.savefig(fn, bbox_inches="tight", pad_inches=0.1)
    print(f"Saved in {fn}")


@handle("gambling")
def gambling():
    pdf = (0.41, 0.17, 0.42)
    # You can write your code here if you want

    raise NotImplementedError()


################################################################################


@handle("plot-em")
def plot_em():
    (X,) = load_dataset("mixture", "X")

    mus = np.linspace(-2, 8, num=1_000)
    # compute log_liks: an array with shape the same as mus containing
    # the log-likelihood for each mu value
    # (ignoring an additive constant is fine)
    raise NotImplementedError()

    em_mus = [mu := 0]
    for _ in range(5):
        mu = np.random.normal()  # TODO: replace
        em_mus.append(mu)
    em_mus = np.asarray(em_mus)

    fig, ax = plt.subplots()
    ax.set_xlabel("$\mu$")
    ax.set_ylabel("log-likelihood")

    ax.plot(mus, log_liks)
    plt.axvline(mus[np.argmax(log_liks)], color="k", ls="--")
    plt.plot(em_mus, log_liks[mus.searchsorted(em_mus)], color="r", ls="--", marker="o")

    ax.set_xlim(mus[0], mus[-1])

    fn = Path("..") / "figs" / "mixture-fitting.pdf"
    fig.savefig(fn, bbox_inches="tight", pad_inches=0.1)
    print(f"Saved in {fn}")


################################################################################


def eval_models(models, dataset="mnist", **kwargs):
    X, y, Xtest, ytest = load_dataset(dataset, "X", "y", "Xtest", "ytest", **kwargs)
    for model in models:
        model.fit(X, y)
        yhat = model.predict(Xtest)
        yield np.mean(yhat != ytest)


def eval_model(model, dataset="mnist", **kwargs):
    return next(eval_models([model], dataset=dataset, **kwargs))


@handle("mnist-nb")
def mnist_nb():
    laps = [1e-9, 0.5, 1, 2.5, 5, 10, 50]
    errs = eval_models(
        [NaiveBayes(prior_alpha=lap, prior_beta=lap) for lap in laps], binarize=True
    )
    print("NaiveBayes test set errors:")
    for lap, err in zip(laps, errs):
        print(f"  lap    {lap:>4.1f}:  {err:.1%}")


@handle("mnist-logreg")
def mnist_logreg():
    from sklearn.linear_model import LogisticRegression

    err = eval_model(
        LogisticRegression(
            solver="saga",
            penalty="l2",
            C=1,  # like 1/lambda
            tol=0.1,  # optimization tolerance
            max_iter=10_000,
            multi_class="multinomial",  # use softmax loss
        )
    )
    print(f"Test set error: {err:.1%}")


@handle("mnist-vqnb")
def mnist_vqnb():
    ks = [2, 3, 4, 5]
    models = [VQNB(k=k) for k in ks]
    print("VQNB test set errors:")
    for k, err in zip(ks, eval_models(models, binarize=True)):
        print(f"  k = {k}:  {err:.1%}")

    model = models[-1]
    fig, axes = plt.subplots(
        k,
        10,
        figsize=(8, 8 * k / 10),
        sharex=True,
        sharey=True,
        constrained_layout=True,
    )
    for y in range(10):
        for b in range(k):
            ps = np.zeros(784)  # get the probabilities from your model
            axes[b][y].imshow(ps.reshape((28, 28)), "gray")
            axes[b][y].set_axis_off()
    fig.savefig("../figs/vqnb_probs.pdf", bbox_inches="tight", pad_inches=0.1)
    print("Plots in ../figs/vqnb_probs.pdf")


################################################################################


@handle("nn-regression")
def nn_regression():
    X, y = load_dataset("basis_data", "X", "y")

    model = NeuralNetRegressor([5], plot_end=True)
    model.fit(X, y)
    yhat = model.predict(X)
    print("Training error: ", np.mean((yhat - y) ** 2))
    print("Figure in ../figs/regression-net.pdf")


@handle("mnist-mlp")
def mnist_mlp():
    import torch_net as tn

    model = tn.TorchNeuralNetClassifier([3], device="cpu")
    # might run faster if you change device= to use your GPU:
    #    "cuda" on Linux/Windows if you have an appropriate GPU and PyTorch install
    #    "mps" if you have a recent Mac
    print(f"mlp test set error: {eval_model(model):.1%}")


@handle("mnist-cnn")
def mnist_cnn():
    import torch_net as tn

    model = tn.Convnet()
    # might run faster if you change device= to use your GPU:
    #    "mps" if you have a recent Mac
    #    "cuda" on Linux/Windows if you have an appropriate GPU and PyTorch install
    print(f"cnn test set error: {eval_model(model):.1%}")


if __name__ == "__main__":
    main()
